from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class AddressConfig(AppConfig):
    """Address application."""

    name = "django_address"
    verbose_name = _("Address")



from django.contrib import admin

from django_address.models import Address, Country, District, Locality, Region, Street


@admin.register(Country)
class CountryAdmin(admin.ModelAdmin):
    search_fields = ("name", "code")


@admin.register(Region)
class RegionAdmin(admin.ModelAdmin):
    search_fields = ("name", "code")


@admin.register(District)
class DistrictAdmin(admin.ModelAdmin):
    search_fields = ("name", "code")


@admin.register(Locality)
class LocalityAdmin(admin.ModelAdmin):
    search_fields = ("name", "postal_code")
    prepopulated_fields = {"slug": ("name",)}


@admin.register(Street)
class StreetAdmin(admin.ModelAdmin):
    list_display = ("name", "locality")
    search_fields = ("name",)


@admin.register(Address)
class AddressAdmin(admin.ModelAdmin):
    search_fields = ("route", "formatted_address")
    list_filter = ["locality"]



import abc
from dataclasses import dataclass
from typing import Union
from uuid import UUID

from django.db import transaction

import swapper


class AddressError(Exception):
    pass


@dataclass
class AbstractAddress(metaclass=abc.ABCMeta):
    """Abstract Class describes address information."""

    raw: str = ""
    country: Union[str, int, UUID] = ""
    country_code: str = ""
    region: Union[str, int, UUID] = ""
    region_code: str = ""
    district: Union[str, int, UUID] = ""
    district_code: str = ""
    locality: Union[str, int, UUID] = ""
    postal_code: str = ""
    street: Union[str, int, UUID] = ""
    street_number: str = ""
    apartment: str = ""
    formatted_address: str = ""
    latitude: float = 0
    longitude: float = 0

    def __post_init__(self):
        self.Country = swapper.load_model("django_address", "Country", required=True)
        self.Region = swapper.load_model("django_address", "Region", required=True)
        self.District = swapper.load_model("django_address", "District", required=True)
        self.Locality = swapper.load_model("django_address", "Locality", required=True)
        self.Street = swapper.load_model("django_address", "Street", required=True)
        self.Address = swapper.load_model("django_address", "Address", required=True)

    @abc.abstractmethod
    def save(self, **kwargs):
        raise NotImplementedError


@dataclass
class Address(AbstractAddress):
    @classmethod
    def _get_or_create(cls, model, value, create=True, **kwargs):
        if isinstance(value, (int, UUID)):
            return model.get_or_none(pk=value)
        obj = None
        if value:
            obj = model.objects.filter(name=value, **kwargs).first()
        return model.objects.create(name=value, **kwargs) if obj is None and create else obj

    def get_or_create_country(self, create=True):
        if isinstance(self.country, self.Country):
            return self.country
        if self.country or self.country_code:
            return self._get_or_create(model=self.Country, value=self.country, create=create, code=self.country_code)
        return None

    def get_or_create_region(self, create=True):
        if isinstance(self.region, self.Region):
            return self.region
        if self.region or self.region_code:
            with transaction.atomic():
                if not isinstance(self.country, self.Country) and create:
                    self.country = self.get_or_create_country()
                return self._get_or_create(
                    model=self.Region, value=self.region, create=create, code=self.region_code, country=self.country
                )
        return None

    def get_or_create_district(self, create=True):
        if isinstance(self.district, self.District):
            return self.district
        if self.district or self.district_code:
            with transaction.atomic():
                if not isinstance(self.region, self.Region) and create:
                    self.region = self.get_or_create_region()
                return self._get_or_create(
                    model=self.District, value=self.district, create=create, code=self.district_code, region=self.region
                )
        return None

    def get_or_create_locality(self, create=True):
        if isinstance(self.locality, self.Locality):
            return self.locality
        if self.locality or self.postal_code:
            with transaction.atomic():
                if not isinstance(self.region, self.Region) and create:
                    self.region = self.get_or_create_region()
                if not isinstance(self.district, self.District) and create:
                    self.district = self.get_or_create_district()
                return self._get_or_create(
                    model=self.Locality,
                    value=self.locality,
                    create=create,
                    postal_code=self.postal_code,
                    region=self.region,
                    district=self.district,
                )
        return None

    def get_or_create_street(self, create=True):
        if isinstance(self.street, self.Street):
            return self.street
        if self.street:
            with transaction.atomic():
                if not isinstance(self.locality, self.Locality) and create:
                    self.locality = self.get_or_create_locality()
                return self._get_or_create(model=self.Street, value=self.street, create=create, locality=self.locality)
        return None

    def save(self):
        """Saves address info to django model."""
        with transaction.atomic():
            try:
                self.street = self.get_or_create_street()
                address, _ = self.Address.objects.get_or_create(
                    locality=self.street.locality,  # noqa
                    street=self.street,
                    raw=self.raw,
                    route=str(self.street),
                    street_number=self.street_number,
                    formatted_address=self.formatted_address,
                    latitude=self.latitude,
                    longitude=self.longitude,
                    apartment=self.apartment,
                )
                return address
            except Exception as error:
                raise AddressError from error



import json

from django.db import models
from django.forms.models import model_to_dict
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _

import swapper


class GetOrNoneManager(models.Manager):
    """Adds get_or_none method to objects."""

    def get_or_none(self, **kwargs):
        try:
            return self.get(**kwargs)
        except self.model.DoesNotExist:
            return None


class AbstractCountryModel(models.Model):
    """Abstract country model."""

    name = models.CharField(_("Name"), max_length=50, unique=True)
    code = models.CharField(_("Code"), max_length=2, blank=True, default="")

    objects = GetOrNoneManager()

    class Meta:
        abstract = True
        ordering = ("name",)
        verbose_name = _("Country")
        verbose_name_plural = _("Countries")

    def __str__(self):
        return "{name}".format(name=self.name or self.code)

    def to_dict(self):
        return model_to_dict(self, fields=[field.name for field in self._meta.fields])


class AbstractAdministrativeAreaLevel1Model(models.Model):
    """Abstract administrative area level 1 model (example: State or Region)."""

    name = models.CharField(_("Name"), max_length=150)
    code = models.CharField(_("Code"), max_length=5, blank=True, default="")
    country = models.ForeignKey(
        to=swapper.get_model_name("django_address", "Country"), on_delete=models.PROTECT, verbose_name=_("Country"),
    )

    objects = GetOrNoneManager()

    class Meta:
        abstract = True
        ordering = ("country", "name")
        unique_together = (("name", "country"),)

    def __str__(self):
        country = "{country}".format(country=self.country)
        name = "{name}".format(name=self.name or self.code)
        if country:
            return "{name}, {country}".format(name=name, country=country)
        return name or country

    def to_dict(self):
        return model_to_dict(self, fields=[field.name for field in self._meta.fields])


class AbstractAdministrativeAreaLevel2Model(models.Model):
    """Abstract administrative area level 2 model."""

    name = models.CharField(_("Name"), max_length=150)
    code = models.CharField(_("Code"), max_length=5, blank=True, default="")

    objects = GetOrNoneManager()

    class Meta:
        abstract = True
        ordering = ("name",)

    def __str__(self):
        suffix = "{suffix}".format(
            suffix=(
                getattr(self, "region", "")
                or getattr(self, "state", "")
                or getattr(self, "province", "")
                or getattr(self, "canton", "")
            )
        )
        name = "{name}".format(name=self.name or self.code)
        if suffix:
            return "{name}, {suffix}".format(name=name, suffix=suffix)
        return name

    def to_dict(self):
        return model_to_dict(self, fields=[field.name for field in self._meta.fields])


class AbstractLocalityModel(models.Model):
    """Abstract locality model."""

    name = models.CharField(_("Name"), max_length=100)
    postal_code = models.CharField(_("Postal code"), max_length=10, blank=True, default="")

    objects = GetOrNoneManager()

    class Meta:
        abstract = True
        ordering = ("name",)
        verbose_name = _("Locality")
        verbose_name_plural = _("Localities")

    def __str__(self):
        suffix = "{suffix}".format(suffix=getattr(self, "district", "") or getattr(self, "county", ""))
        locality = "{name}".format(name=self.name)
        if suffix:
            return "{locality}, {suffix}".format(locality=locality, suffix=suffix)
        return locality

    def to_dict(self):
        return model_to_dict(self, fields=[field.name for field in self._meta.fields])


class AbstractStreetModel(models.Model):
    """Abstract street model."""

    locality = models.ForeignKey(
        to=swapper.get_model_name("django_address", "Locality"), on_delete=models.CASCADE, verbose_name=_("Locality")
    )
    name = models.CharField(_("Street name"), max_length=256)

    objects = GetOrNoneManager()

    class Meta:
        abstract = True
        ordering = ("name",)
        unique_together = (("locality", "name"),)
        verbose_name = _("Street")
        verbose_name_plural = _("Streets")

    def __str__(self):
        return "{name}".format(name=self.name)

    def to_dict(self):
        return model_to_dict(self, fields=[field.name for field in self._meta.fields])


class AbstractAddressModel(models.Model):
    """Abstract address model."""

    locality = models.ForeignKey(
        to=swapper.get_model_name("django_address", "Locality"),
        on_delete=models.CASCADE,
        verbose_name=_("Locality"),
        null=True,
    )
    street = models.ForeignKey(
        to=swapper.get_model_name("django_address", "Street"),
        on_delete=models.CASCADE,
        verbose_name=_("Street"),
        null=True,
    )
    raw = models.CharField(_("Raw"), max_length=256, blank=True, default="")
    route = models.CharField(_("Route"), max_length=256, default="")
    street_number = models.CharField(_("Street number"), max_length=20, blank=True, default="")
    formatted_address = models.CharField(_("Formatted address"), max_length=200, blank=True, default="")
    latitude = models.FloatField(_("Latitude"), blank=True, default=0)
    longitude = models.FloatField(_("Longitude"), blank=True, default=0)
    apartment = models.CharField(_("Apartment"), max_length=10, blank=True, default="")

    objects = GetOrNoneManager()

    class Meta:
        abstract = True
        verbose_name = _("Address")
        verbose_name_plural = _("Addresses")

    def __str__(self):
        if self.formatted_address:
            return self.formatted_address

        locality = "{locality}".format(locality=self.locality)
        route = self.street.name if self.street else self.route
        address = "{route}{street_number}{apartment}".format(
            route=route,
            street_number=", {street_number}".format(street_number=self.street_number) if self.street_number else "",
            apartment=", {apartment}".format(apartment=self.apartment) if self.apartment else "",
        )
        if route and locality:
            return "{address}, {locality}".format(address=address, locality=locality)
        if route:
            return address

        return self.raw

    def save(self, *args, **kwargs):
        if not self.route and self.street:
            self.route = str(self.street)
        if not self.locality and self.street:
            self.locality = self.street.locality
        if not self.formatted_address:
            self.formatted_address = str(self)
        return super().save(*args, **kwargs)

    def to_dict(self):
        return {
            "raw": self.raw,
            "locality": self.locality.to_dict() if self.locality else "",
            "street": self.street.to_dict() if self.street else "",
            "route": self.route,
            "street_number": self.street_number,
            "latitude": self.latitude,
            "longitude": self.longitude,
            "formatted_address": self.formatted_address,
        }

    def to_json(self):
        return json.dumps(self.to_dict())


class Country(AbstractCountryModel):
    """Country model."""

    class Meta(AbstractCountryModel.Meta):
        swappable = swapper.swappable_setting("django_address", "Country")


class Region(AbstractAdministrativeAreaLevel1Model):
    """Region model."""

    class Meta(AbstractAdministrativeAreaLevel1Model.Meta):
        swappable = swapper.swappable_setting("django_address", "Region")
        verbose_name = _("Region")
        verbose_name_plural = _("Regions")


class District(AbstractAdministrativeAreaLevel2Model):
    """District model."""

    region = models.ForeignKey(
        to="Region",
        on_delete=models.CASCADE,
        verbose_name=_("Region"),
        related_name="districts",
    )

    class Meta(AbstractAdministrativeAreaLevel2Model.Meta):
        swappable = swapper.swappable_setting("django_address", "District")
        ordering = ("region", "name")
        unique_together = (("name", "region"),)
        verbose_name = _("District")
        verbose_name_plural = _("Districts")


class Locality(AbstractLocalityModel):
    """Locality model."""

    slug = models.SlugField(_("Slug"), max_length=100, unique=True)
    region = models.ForeignKey(
        to="Region",
        on_delete=models.CASCADE,
        verbose_name=_("Region"),
        related_name="localities",
    )

    district = models.ForeignKey(
        to="District",
        on_delete=models.CASCADE,
        verbose_name=_("District"),
        related_name="localities",
        blank=True,
        null=True,
    )

    class Meta(AbstractLocalityModel.Meta):
        swappable = swapper.swappable_setting("django_address", "Locality")
        ordering = ("region", "district", "name")

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.name)
        super().save(*args, **kwargs)


class Street(AbstractStreetModel):
    """Street model."""

    class Meta(AbstractStreetModel.Meta):
        swappable = swapper.swappable_setting("django_address", "Street")


class Address(AbstractAddressModel):
    """Address model."""

    class Meta(AbstractAddressModel.Meta):
        swappable = swapper.swappable_setting("django_address", "Address")

    def to_dict(self):
        address = {
            "raw": self.raw,
            "locality": self.locality.to_dict() if self.locality else "",
            "street": self.street.to_dict() if self.street else "",
            "route": self.route,
            "street_number": self.street_number,
            "latitude": self.latitude,
            "longitude": self.longitude,
            "formatted_address": self.formatted_address,
        }

        district = self.locality.district if self.locality else None
        region = self.locality.region if self.locality else None
        country = region if region else None
        if country:
            address.update({"country": country.to_dict()})
        if region:
            address.update({"region": region.to_dict()})
        if district:
            address.update({"district": district.to_dict()})
        return address



import django.db.models.deletion
from django.db import migrations, models

import swapper


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Country',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50, unique=True, verbose_name='Name')),
                ('code', models.CharField(blank=True, default='', max_length=2, verbose_name='Code')),
            ],
            options={
                'verbose_name': 'Country',
                'verbose_name_plural': 'Countries',
                'ordering': ('name',),
                'abstract': False,
                'swappable': swapper.swappable_setting("django_address", "Country"),
            },
        ),
        migrations.CreateModel(
            name='District',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, verbose_name='Name')),
                ('code', models.CharField(blank=True, default='', max_length=5, verbose_name='Code')),
            ],
            options={
                'verbose_name': 'District',
                'verbose_name_plural': 'Districts',
                'ordering': ('region', 'name'),
                'abstract': False,
                'swappable': swapper.swappable_setting("django_address", "District"),
            },
        ),
        migrations.CreateModel(
            name='Locality',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.SlugField(max_length=100, unique=True, verbose_name='Slug')),
                ('name', models.CharField(max_length=100, verbose_name='Name')),
                ('postal_code', models.CharField(blank=True, default='', max_length=10, verbose_name='Postal code')),
                ('district', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='localities', to="django_address.District", verbose_name='District')),
            ],
            options={
                'verbose_name': 'Locality',
                'verbose_name_plural': 'Localities',
                'ordering': ('region', 'district', 'name'),
                'abstract': False,
                'swappable': swapper.swappable_setting("django_address", "Locality"),
            },
        ),
        migrations.CreateModel(
            name='Street',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256, verbose_name='Street name')),
                ('locality', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="django_address.Locality", verbose_name='Locality')),
            ],
            options={
                'verbose_name': 'Street',
                'verbose_name_plural': 'Streets',
                'ordering': ('name',),
                'abstract': False,
                'swappable': swapper.swappable_setting("django_address", "Street"),
                'unique_together': {('locality', 'name')},
            },
        ),
        migrations.CreateModel(
            name='Region',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, verbose_name='Name')),
                ('code', models.CharField(blank=True, default='', max_length=5, verbose_name='Code')),
                ('country', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to="django_address.Country", verbose_name='Country')),
            ],
            options={
                'verbose_name': 'Region',
                'verbose_name_plural': 'Regions',
                'ordering': ('country', 'name'),
                'abstract': False,
                'swappable': swapper.swappable_setting("django_address", "REGION"),
                'unique_together': {('name', 'country')},
            },
        ),
        migrations.AddField(
            model_name='locality',
            name='region',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='localities', to="django_address.Region", verbose_name='Region'),
        ),
        migrations.AddField(
            model_name='district',
            name='region',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='districts', to="django_address.Region", verbose_name='Region'),
        ),
        migrations.CreateModel(
            name='Address',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('raw', models.CharField(blank=True, default='', max_length=256, verbose_name='Raw')),
                ('route', models.CharField(default='', max_length=256, verbose_name='Route')),
                ('street_number', models.CharField(blank=True, default='', max_length=20, verbose_name='Street number')),
                ('formatted_address', models.CharField(blank=True, default='', max_length=200, verbose_name='Formatted address')),
                ('latitude', models.FloatField(blank=True, default=0, verbose_name='Latitude')),
                ('longitude', models.FloatField(blank=True, default=0, verbose_name='Longitude')),
                ('apartment', models.CharField(blank=True, default='', max_length=10, verbose_name='Apartment')),
                ('locality', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to="django_address.Locality", verbose_name='Locality')),
                ('street', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to="django_address.Street", verbose_name='Street')),
            ],
            options={
                'verbose_name': 'Address',
                'verbose_name_plural': 'Addresses',
                'abstract': False,
                'swappable': swapper.swappable_setting("django_address", "Address"),
            },
        ),
        migrations.AlterUniqueTogether(
            name='district',
            unique_together={('name', 'region')},
        ),
    ]



from uuid import UUID

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.fields.related_descriptors import ForwardManyToOneDescriptor
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

import swapper

from django_address.service import AddressError


class AddressDescriptor(ForwardManyToOneDescriptor):
    def __set__(self, inst, value):
        self.Address = swapper.load_model("django_address", "Address", required=True)  # noqa
        super().__set__(inst, self.to_python(value))

    def to_python(self, value):

        if value is None:
            return None

        if isinstance(value, (self.Address, int, UUID)):
            return value

        if isinstance(value, dict):
            try:
                address_svc = import_string(
                    getattr(settings, "DJANGO_ADDRESS_SERVICE_CLASS", "django_address.service.Address")
                )
                new_address = address_svc(**value)
                return new_address.save()
            except AddressError:
                raise ValidationError("Invalid address value.")

        raise ValidationError("Invalid address value.")


class AddressField(models.ForeignKey):
    """Address field."""

    description = _("Address")

    def __init__(self, *args, **kwargs):
        kwargs["to"] = swapper.get_model_name("django_address", "Address")
        kwargs["on_delete"] = models.CASCADE
        super(AddressField, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name, virtual_only=False, **kwargs):
        super().contribute_to_class(cls, name, private_only=virtual_only, **kwargs)
        setattr(cls, self.name, AddressDescriptor(self))



from setuptools import find_packages, setup

from django_address import __version__ as version

requirements = ["Django>=2.2", "requests", "typing-extensions", "psycopg2-binary", "environs", "structlog", "swapper"]

extras_require = {
    "test": ["pytest-cov", "pytest-django", "pytest"],
    "lint": ["flake8", "wemake-python-styleguide", "isort"],
}

extras_require["dev"] = extras_require["test"] + extras_require["lint"]  # noqa: W504  # noqa: W504

with open("README.md", "r", encoding="utf-8") as readme:
    long_description = readme.read()


setup(
    name="django-address-app",
    author="Vyacheslav Onufrienko",
    author_email="onufrienkovi@gmail.com",
    description="Django models for storing and retrieving postal addresses.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=version,
    url="https://github.com/onufrienkovi/django-address-app",
    extras_require=extras_require,
    packages=find_packages(exclude=["tests", "docs", "scripts", "example"]),
    install_requires=requirements,
    python_requires=">=3.7",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
